<?php

namespace App\Console\Commands;

use _PHPStan_993c0a2e7\RingCentral\Psr7\LazyOpenStream;
use App\Exceptions\KeycloakException;
use App\Models\Databases;
use App\Services\DirectAccess;
use App\Services\Keycloak;
use App\Services\VVInterface;
use Illuminate\Console\Command;
use Illuminate\Support\Facades\Log;

/**
 * Command for synchronizing VV users with Keycloak.
 * Format: php artisan keycloak:sync {mode=hourly} {club?} {operation?}
 *    mode: hourly (default), daily, now, reverse
 *    club: database id and club joined with an underscore, e.g. 239_28994
 *    operation: enable, disable, delete
 */
class KeycloakSync extends Command
{
    /**
     * The name and signature of the console command.
     *
     * @var string
     */
    protected $signature = 'keycloak:sync {mode=hourly} {club?} {operation?}';

    /**
     * The console command description.
     *
     * @var string
     */
    protected $description = 'Synchronize keycloak users with VV.';

    /**
     * Create a new command instance.
     *
     * @return void
     */
    public function __construct()
    {
        parent::__construct();
    }

    /**
     * Execute the console command.
     *
     * @return mixed
     * @throws KeycloakException
     */
    public function handle( Keycloak $keycloak )
    {
        $return = true;
        $mode = $this->argument('mode');
        if( !in_array(strtolower($mode), ['hourly', 'daily', 'now', 'reverse']) ) {
            throw new KeycloakException("Invalid mode $mode");
        }

        try {
            switch( $mode ) {
                case 'now':
                    $return = $this->handleNow();
                    break;
                case 'reverse':
                    $return = $keycloak->synchronizeReverse();
                    break;
                default:
                    $keycloak->synchronize($mode);
            }
        } catch( KeycloakException $e ) {
            $return = false;
            Log::debug( "KeycloackSync::handle Error (mode $mode): " . $e->getMessage() );
            $this->error( $e->getMessage() );
        }

        return $return;
    }

    /**
     * Handle the 'now' mode.
     *
     * @return boolean True if the operation was successful, false otherwise.
     * @throws KeycloakException
     */
    private function handleNow()
    {
        // club is mandatory for now mode
        $club = $this->argument('club');
        if( empty($club) ) {
            throw new KeycloakException("Club is required for mode 'now'");
        }
        elseif( !preg_match('/^\d+_\d+$/', $club) ) {
            throw new KeycloakException("Invalid club $club");
        }

        // operation is mandatory for now mode
        $operation = $this->argument('operation');
        if( empty($operation) ) {
            throw new KeycloakException("Operation is required for mode 'now'");
        }
        if( !in_array(strtolower($operation), ['enable', 'disable', 'delete']) ) {
            throw new KeycloakException("Invalid operation $operation");
        }
        $operation = strtoupper($operation);

        // Split database id and club id
        list($database_id, $club_id) = explode('_', $club);
        // Validate database_id
        $dbdata = Databases::getIfexists($database_id);
        if( $dbdata === FALSE ) {
            abort( 400, 'Invalid database id' );
        }

        $vvClient = new DirectAccess($dbdata);

        try {
            $vvClient->synchronizeNow($club_id, $operation);
        } catch( KeycloakException $e ) {
            Log::debug( "KeycloackSync::handle Error: " . $e->getMessage() );
            $this->error( $e->getMessage() );
            return false;
        }

        $vvClient->vvClose();
        return true;
    }
}
