<?php
namespace App\Services;

use App\Jobs\CancelClubs;
use App\Jobs\ChangeClubs;
use App\Jobs\DeleteClubs;
use App\Jobs\UpDowngradeClubs;
use App\Models\Delayed;
use Illuminate\Support\Facades\Log;

/**
 * Helper class to handle delayed actions
 */
class DelayedActions {

    const ACTION_TYPE_DELETE = 1;
    const ACTION_TYPE_UPDOWNGRADE = 2;
    const ACTION_TYPE_CANCEL = 3;
    const ACTION_TYPE_CHANGE = 4;

    /**
     * Create a new delayed action.
     *
     * @param int $p_iDatabaseId Database ID
     * @param int $p_iClubId Unique club ID
     * @param int $p_iType Type (one of ACTION_TYPE_*)
     * @param int $p_iActionDate Date of action in UNIX timestamp format
     */
    public static function create($p_iDatabaseId, $p_iClubId, $p_iType, $p_iActionDate, $p_bNewState = null ) {
      // Record data in helper table for delete monitoring
      $delayedAction = new Delayed();
      $delayedAction->database_id = $p_iDatabaseId;
      $delayedAction->club_id = $p_iClubId;
      $delayedAction->type = $p_iType;
      $delayedAction->action_date = $p_iActionDate;
      if( !is_null($p_bNewState) ) {
        $delayedAction->new_state = ($p_bNewState ? 1 : 0);
      }
      $delayedAction->save();
    }

    /**
     * Check the Delayed table for necessary actions based on their action date and execute them.
     * This is called by the scheduler.
     *
     * @return void
     * @throws \Exception
     */
   public function __invoke()
   {
       $delayedActions = (new Delayed())->currentActions();

       foreach ($delayedActions as $action) {
           switch ($action->type) {
               case DelayedActions::ACTION_TYPE_DELETE:
                   $job = (new DeleteClubs($action->database_id, $action->club_id))->onQueue('delete');
                   dispatch($job);
                   (new Delayed())->where('id', $action->id)->delete();
                   break;
               case DelayedActions::ACTION_TYPE_UPDOWNGRADE:
                   $job = (new UpDowngradeClubs($action->database_id, $action->club_id, $action->new_state))->onQueue('updowngrade');
                   dispatch($job);
                   (new Delayed())->where('id', $action->id)->delete();
                   break;
               case DelayedActions::ACTION_TYPE_CANCEL:
                   $job = (new CancelClubs($action->database_id, $action->club_id))->onQueue('cancel');
                   dispatch($job);
                   (new Delayed())->where('id', $action->id)->delete();
                   break;
               case DelayedActions::ACTION_TYPE_CHANGE:
                   $job = (new ChangeClubs($action->database_id, $action->club_id, $action->new_state))->onQueue('change');
                   dispatch($job);
                   (new Delayed())->where('id', $action->id)->delete();
                   break;
               default:
                   throw new \Exception('Unknown action type');
           }
       }
   }

}

