<?php

namespace App\Services;

use App\Exceptions\KeycloakException;
use App\Models\Adressen;
use App\Models\Information;
use App\Models\MIOKunden;
use Carbon\Carbon;
use DateTime;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Config;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Log;

/**
 * This class handles direct access to the VV databases for some statistical information.
 * This was implemented to avoid complications with automatic database configuration on the
 * VVInterface class, which made that approach impractical.  This class handles both Oracle and
 * MySQL queries.
 */
class DirectAccess
{
    const USE_CONNECTION = 'vvclient';
    private $m_strDefaultConnection;
    private $m_oDefaultConnectionData;
    private $m_aDBConnectionParams;

    /**
     * Class constructor.
     * Changes de database connection environment to the one passed in the array.
     *
     * @param array $p_aDBData Array with database connection parameters.
     */
    public function __construct( $p_aDBData )
    {
        if( !is_array($p_aDBData) && is_object($p_aDBData) ) {
            $p_aDBData = $p_aDBData->toArray();
        }

        $this->m_aDBConnectionParams = $p_aDBData;
        $this->m_strDefaultConnection = DB::getDefaultConnection();
        DB::setDefaultConnection(self::USE_CONNECTION );

        $this->m_oDefaultConnectionData = Config::get('database.connections.vvclient');

        $clientConnection = array();
        $clientConnection["driver"] = $p_aDBData["dbtype"];
        $clientConnection["host"] = $p_aDBData["dbhost"];
        $clientConnection["port"] = $p_aDBData["dbport"];
        $clientConnection["username"] = $p_aDBData["dbuser"];
        $clientConnection["password"] = $p_aDBData["dbuserpassword"];

        if( $p_aDBData["dbtype"] == "oracle" ) {
            $clientConnection["service_name"] = $p_aDBData["dbname"];
            $clientConnection["database"] = "";
            $clientConnection["charset"] = "AL32UTF8";
        }
        else {
            $clientConnection["service_name"] = "";
            $clientConnection["database"] = $p_aDBData["dbname"];
            $clientConnection["charset"] = "utf8mb4";
        }

        Config::set('database.connections.vvclient', $clientConnection);

        DB::purge( self::USE_CONNECTION );
        DB::reconnect( self::USE_CONNECTION );
    }

    /**
     * Destructor.
     * Restores the original database configuration.
     */
    public function __destruct()
    {
        // Reset DB connection
        Config::set('database.connections.vvclient', $this->m_oDefaultConnectionData);
        DB::purge( self::USE_CONNECTION );
        DB::reconnect( self::USE_CONNECTION );

        DB::setDefaultConnection($this->m_strDefaultConnection);
    }

    public function vvClose()
    {
        $this->__destruct();
    }

    /**
     * Obtain database size for an Oracle database.
     * Requires a SYSTEM user connection to be set previously via the constructor.
     *
     * @return integer Database size
     * @throws \Exception when calling for a non-Oracle database
     */
    public function vvGetDBSize()
    {
        $iSize = 0;

        if( $this->m_aDBConnectionParams["dbtype"] != "oracle" ) {
            throw new \Exception("This call is only allowed for Oracle databases");
        }

        $results = DB::select("SELECT ROUND(SUM(BYTES)/1024/1024,2) TOT_MB ".
                                "FROM v\$datafile WHERE NAME NOT LIKE '%/undo%'" );
        if( $results !== FALSE ) {
            $iSize = floor($results[0]->tot_mb);
        }

        return $iSize;
    }

    /**
     * Get the total number of clubs.
     *
     * @return integer Number of clubs in the database.
     */
    public function vvGetTotalClubs()
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)->count();
    }

    /**
     * Get the number of active clubs.
     *
     * @return integer Number of active clubs in the database.
     */
    public function vvGetActiveClubs()
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)->where("active", "=", 1 )->count();
    }

    /**
     * Get the number of inactive clubs.
     *
     * @return integer Number of inactive clubs in the database.
     */
    public function vvGetInactiveClubs()
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)->where("active", "=", 0 )->count();
    }

    /**
     * Get the number of trial clubs.
     *
     * @return integer Number of trial clubs in the database.
     */
    public function vvGetTrialClubs()
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        $iCnt = $mioKd->on(self::USE_CONNECTION)
            ->leftJoin( "INFORMATION", function($join) {
                $join->on("INFORMATION.mioid", "=", "MIOKUNDEN.mioid")
                    ->where("INFORMATION.typeid", "=", "2")
                    ->where("INFORMATION.label", "=", "EndDate");
            })
            ->whereNotNull("INFORMATION.content")->count();

        return $iCnt;
    }

    /**
     * Get the club id (kundeId) of the passed club by its shortkey.
     *
     * @param string $p_strShortkey Shortkey used to find club
     * @return integer KundeID/MIOId of club identified by passed shortkey.
     */
    public function vvGetKundeId($p_strShortkey)
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)->where("shortkey", "=", $p_strShortkey )->value('mioid');
    }

    /**
     * Get the shortkey from the club id
     *
     * @param integer $p_iKundeId ID of club
     * @return string Shortkey of corresponding club
     */
    public function vvGetShortkey($p_iKundeId)
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)->where("mioid", "=", $p_iKundeId )->value('shortkey');
    }

    /**
     * Check if the given KundeId exists in the database.
     *
     * @param integer $p_iKundeId ID of club
     * @return bool true if the club exists, false otherwise
     */
    public function vvClubExists($p_iKundeId)
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)->where("mioid", "=", $p_iKundeId )->exists();
    }

    /**
     * Get list of all clubs in the current database.
     *
     * @return \Illuminate\Database\Eloquent\Builder[]|\Illuminate\Database\Eloquent\Collection
     */
    public function vvGetAllClubs()
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        return $mioKd->on(self::USE_CONNECTION)
            ->select("name", "mioid", "shortkey", "productvariant" )
            ->get()->toArray();
    }

    /**
     * Get all the data of the specified club.
     * @param int $p_iKundeId ID of the club
     * @return \Illuminate\Database\Eloquent\Builder[]|\Illuminate\Database\Eloquent\Collection
     */
    public function vvClubData($p_iKundeId)
    {
        $mioKd = new MIOKunden();
        $mioKd->refresh();
        $data = $mioKd->on(self::USE_CONNECTION)
            ->where("mioid", "=", $p_iKundeId)
            ->first();
        if( !$data->isEmpty() ) {
            return $data->toArray();
        }
        else {
            return array();
        }
    }

    /**
     * Get list of members of the given club with their Keycloak state info.
     *
     * @param int $club_id
     * @param array $states
     * @param bool $p_bOnlyActive
     * @return \Illuminate\Database\Eloquent\Builder[]|\Illuminate\Database\Eloquent\Collection
     */
    public function vvGetMembersWithKCState(int $club_id, array $states, bool $p_bOnlyActive = false)
    {
        //Log::debug("vvGetMembersWithKCState: ".print_r($states, true));
        $adresse = new Adressen();
        $adresse->refresh();
        $adresse = $adresse->on(self::USE_CONNECTION)
            ->from("ADRESSEN as a")
            ->select("a.adrid", "a.nachname", "a.vorname", "a.keycloakemail", "a.keycloakstate",
                "a.keycloakrole", "a.keycloakdate", "a.keycloakuserid", "a.keycloakuser",
                "g.geburtsdatum")
            ->join("GESCHAEFTSINFO as g", function($join) {
                $join->on("g.mioid", "=", "a.mioid");
                $join->on("g.adrid", "=", "a.adrid");
            })
            ->where("a.mioid", "=", $club_id)
            ->where("a.adresstypid", "=", 2)
            ->whereIn("a.keycloakstate", $states);
        if( $p_bOnlyActive ) {
            $adresse = $adresse
                ->whereNull( "g.austrittsdatum" )
                ->orWhere( "g.austrittsdatum", ">=", (new DateTime('00:00'))->format('Y.m.d H:i:s'));
        }

        return $adresse->get();
    }

    /**
     * Save Keycloakdata in given member record.
     *
     * @param int $kundeId
     * @param int $adrId
     * @param array $kckData
     * @return boolean true on success, false on failure
     */
    public function vvSaveKCData( int $kundeId, int $adrId, array $kckData )
    {
        //Log::debug("vvSaveKCData: Kd=$kundeId, Adr=$adrId, ".print_r($kckData, true));
        $aUpdateData['keycloakstate'] = $kckData["state"];
        if( isset($kckData["userid"]) ) {
            $aUpdateData['keycloakuserid'] = $kckData["userid"];
        }
        if( isset($kckData["email"]) ) {
            $aUpdateData['keycloakemail'] = $kckData["email"];
        }
        if( isset($kckData["user"]) ) {
            $aUpdateData['keycloakuser'] = $kckData["user"];
        }
        if( isset($kckData["role"]) ) {
            $aUpdateData['keycloakrole'] = $kckData["role"];
        }
        if( isset($kckData["date"]) ) {
            $aUpdateData['keycloakdate'] = new Carbon($kckData["date"]);
        }
        else {
            $aUpdateData['keycloakdate'] = Carbon::now();
        }
        $rt = Adressen::on(self::USE_CONNECTION)
            ->where('mioid', $kundeId)
            ->where('adrid', $adrId)
            ->update($aUpdateData);
        //Log::debug("vvSaveKCData: Kd=$kundeId, Adr=$adrId, updateData: ".print_r($aUpdateData, true) . ", RT=$rt");
        return ($rt > 0);
    }

    /**
     * Delete Keycloakdata in given member record.
     *
     * @param int $kundeId KundeId of the club
     * @param int $adrId AdrId of the member
     * @return boolean true on success, false on failure
     */
    public function vvDeleteKCData(int $kundeId, int $adrId )
    {
        $aUpdateData['keycloakstate'] = null;
        $aUpdateData['keycloakuserid'] = null;
        $aUpdateData['keycloakemail'] = null;
        $aUpdateData['keycloakuser'] = null;
        $aUpdateData['keycloakrole'] = null;
        $aUpdateData['keycloakdate'] = null;

        $rt = Adressen::on(self::USE_CONNECTION)
            ->where('mioid', $kundeId)
            ->where('adrid', $adrId)
            ->update($aUpdateData);

        return ($rt > 0);
    }

    /**
     * @param int $kundeId
     * @param int $typeId
     * @return \Illuminate\Database\Eloquent\Builder[]|\Illuminate\Database\Eloquent\Collection
     */
    public function vvGetInfoByType(int $kundeId, int $typeId)
    {
        $info = new Information();
        $info->refresh();
        return $info->on(self::USE_CONNECTION)
            ->where("mioid", "=", $kundeId)
            ->where("typeid", "=", $typeId)
            ->orderBy("sort")
            ->get()->toArray();
    }

    /**
     * @param int $kundeId
     * @param int $typeId
     * @return mixed
     */
    public function vvDeleteInfoByType(int $kundeId, int $typeId)
    {
        $info = new Information();
        $info->refresh();
        return $info->on(self::USE_CONNECTION)
            ->where("mioid", "=", $kundeId)
            ->where("typeid", "=", $typeId)
            ->delete();
    }

    /**
     * Check if the passed product is contained in the product variant setting of the club.
     * First byte from right to left is App, second is SMALL/LARGE, third is FIBU Premium.
     * Note that FIBU Light (the DFB application) is not handled in ProductVariant but
     * in a separate MIOKUNDEN field "DFBnetFibu".
     * Important note about APP: Asking for APP will also check for APP Light and return true
     * if either APP or APP Light is set. This is to avoid having to ask for both everywhere
     * when only enabling menu items or features.  If special behavior is required for APP or
     * APP Light, check for APP Light first.
     *
     * For clarification, possible values:
     *   0 = 000 = NULL: LARGE           reverted 000
     *   001 = 01: LARGE + APP                    100
     *   002 = 02: LARGE + APP Light              200
     *   010 = 10: SMALL                          010
     *   011 = 11: SMALL + APP                    110
     *   012 = 12: SMALL + APP Light              210
     *   100: LARGE + FIBU Premium                001
     *   110: SMALL + FIBU Premium                011
     *   111: SMALL + FIBU Premium + App          111
     *   112: SMALL + FIBU Premium + App Light    211
     *   101: LARGE + FIBU Premium + App          101
     *   102: LARGE + FIBU Premium + App Light    201
     *
     * @param int $p_iProdVariant Content of field ProductVariant of MIOKunden table
     * @param string $p_strProduct Product to check for (one of the VV_PRODUCT_* constants)
     * @return bool TRUE of product is present, FALSE otherwise.
     */
    private function ProductVariantHas( $p_iProdVariant, $p_strProduct )
    {
      $iCurProdVar = $p_iProdVariant;
      if( is_null($iCurProdVar) ) {
        $iCurProdVar = 0;
      }
      $strCurProd = strrev( sprintf( "%03d", $iCurProdVar ) );  // add leading zeros and revert string, 100 -> 001
      $aProducts = str_split($strCurProd);

      switch ($p_strProduct) {
        case config('constants.VV.PRODUCT_APP'):
          return $aProducts[0] == "1";
        case config('constants.VV.PRODUCT_APP_LIGHT'):
          return ($aProducts[0] == "2");
        case config('constants.VV.PRODUCT_SMALL'):
          return ($aProducts[1] == "1");
        case config('constants.VV.PRODUCT_LARGE'):
          return ($aProducts[1] == "0");
        case config('constants.VV.FIBU_PREMIUM'):
          return ($aProducts[2] == "1");
      }

      return FALSE;
    }

    /**
     * Synchronizes Keycloak user data with the local database on a daily basis.
     * Basically batch processes all users that were modified in the local database
     * and updates their state towards Keycloak (so accesses Keycloak also in write mode).
     * This is used to enable, disable or delete users from Keycloak when the club's
     * licensing state changes or a member is deleted.
     *
     * This function performs the following steps:
     * 1. Retrieves all clubs from the database.
     * 2. Iterates through each club and processes only those with app or app light.
     * 3. Depending on the club's app user state, retrieves members and updates their Keycloak status.
     * 4. Handles different licensing states (enable, disable, delete) and updates the local database accordingly.
     * 5. Logs errors and statistics for the synchronization process.
     *
     * In order to avoid overloading the Keycloak server, the function waits for a few seconds after each
     * batch of requests.
     *
     * @return int|void Returns -1 on configuration failure, 1 if no data is found, or void on successful synchronization.
     * @throws KeycloakException If there is an error during the synchronization process.
     */
    public function synchronizeDaily()
    {
        $aStatistics = $this->defineStatistics(config('constants.keycloakSync.MODE_DAILY'));
        $aStates = $this->defineStates(config('constants.keycloakSync.MODE_DAILY'));

        if (!Keycloak::setConfiguration(config('keycloak_api.keyCloakConfig'))['ok']) {
            return -1;
        }

        Keycloak::setTime((new DateTime('now'))->format('Y.m.d.H.i.s'));

        $oAllClubs = $this->vvGetAllClubs();
        if( empty($oAllClubs) ) {
            // No data
            return 1;
        }

        foreach ($oAllClubs as $Kunde) {
            // Only process clubs with app or app light
            if( !$this->ProductVariantHas( $Kunde["productvariant"], config('constants.VV.PRODUCT_APP') ) &&
                !$this->ProductVariantHas( $Kunde["productvariant"], config('constants.VV.PRODUCT_APP_LIGHT') ) ) {
                continue;
            }

            $aStatistics['ERR-Process-Keycloak-Failed'][$Kunde["mioid"]] = array();

            //THE DEFAULT TARGET FOR THIS PROCESS IS TO UPDATE RETIRED USERS.
            $bOnlyActive = true;
            $club_app_users_state = 'club_app_users_state_none';
            $aStateInfo = $this->vvGetInfoByType( Information::OPHINEX_INFO_TYPE_APP_USERS_STATE, $Kunde["mioid"] );

            if(!empty($aStateInfo)) {
               // HERE WE SET THE TARGET OF THE PROCESS ACCORDING
               // SOME ACTION SET FOR THE CLUB DUE TO LICENSING CHANGES
               // THIS STATES WILL TRIGGER MASSIVE UPDATES.
               switch($aStateInfo[0]['Content']){
                   case 'ENABLE':
                       $bOnlyActive = true;
                       $club_app_users_state = 'club_app_users_state_enable';
                       break;
                   case 'DISABLE':
                       $bOnlyActive = false;
                       $club_app_users_state = 'club_app_users_state_disable';
                       break;
                   case 'DELETE':
                       $bOnlyActive = false;
                       $club_app_users_state = 'club_app_users_state_delete';
                       break;
               }
            }

            try {
                $oMembers = $this->vvGetMembersWithKCState($Kunde["mioid"], $aStates[$club_app_users_state], $bOnlyActive);
                if (empty($oMembers)) {
                    continue;
                }
            }
            catch( \Exception $e ) {
                Log::error( "Error getting members for club {$Kunde["mioid"]} {$Kunde["name"]}: " . $e->getMessage() );
                continue;
            }

            $iCounter = 0;
            foreach ($oMembers as $member) {
                if (($club_app_users_state == 'club_app_users_state_none') &&
                    (!isset($member->keycloakuserid) || trim($member->keycloakuserid) === '')) {
                    continue;
                }

                if( $iCounter % Keycloak::KEYCLOAKAPI_BATCH_SIZE == 0 ) {
                  // Wait for a few seconds before sending more requests to Keycloak
                  sleep(Keycloak::KEYCLOAKAPI_BATCH_PAUSE );
                }

                switch($club_app_users_state) {
                    case 'club_app_users_state_none':
                        // THERE IS NO CHANGE IN LICENSING STATE, THIS IS THE DEFAULT ACTION
                        // The user is simply checked for existance in KC.
                        $return = Keycloak::statusUser( array(
                            'id' => $member->keycloakuseri
                        ));
                        if (!is_array($return) || !isset($return['ok'])) {
                            continue 2;
                        }
                        continue 2;

                    case 'club_app_users_state_enable':
                        //THE STATE OF APP LICENSING IS SET TO ENABLED
                        $iModState = -1;
                        switch($member->keycloakstate) {
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED'):
                                $iModState = config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED');
                                break;
                        }
                        if( $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED') ||
                            $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED') ){
                            $return = Keycloak::enableUser(array('id' => $member->keycloakuserid));
                        }
                        break;

                    case 'club_app_users_state_disable':
                        //THE STATE OF APP LICENSING IS SET TO DISABLED
                        $iModState = -1;
                        switch($member->keycloakstate) {
                            case config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED');
                                break;
                        }

                        if( $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED') ||
                            $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED') ) {
                            $return = Keycloak::disableUser(array('id' => $member->keycloakuserid));
                        }
                        break;

                    case 'club_app_users_state_delete':
                        //THE STATE OF APP LICENSING IS SET TO DELETED
                        $iModState = -1;
                        switch($member->keycloakstate) {
                            case config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_EMAILVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_DISABLED_EMAILNOTVERIFIED');
                                break;
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED'):
                            case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED'):
                                $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_DISABLED_EMAILVERIFIED');
                                break;
                        }
                        if( $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILNOTVERIFIED')          ||
                            $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILVERIFIED')             ||
                            $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILNOTVERIFIED') ||
                            $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILVERIFIED') ){
                            $return = Keycloak::deleteUser(array('id' => $member->keycloakuserid));
                        }
                        break;
                }

                $iCounter++;

                if( ( $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILNOTVERIFIED')          &&
                      $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILVERIFIED')             &&
                      $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILNOTVERIFIED') &&
                      $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILVERIFIED') ) ||
                    ( isset($return) && is_array($return) && isset($return['ok']) && $return['ok'] === config('constants.keycloakSync.KEYCLOAKAPI_SUCCESS') )
                  ) {
                    $this->vvSaveKCData($Kunde["mioid"], $member->adrid,
                                   array( 'userid' => (($club_app_users_state !== 'club_app_users_state_delete') ? $member->keycloakuserid:'NULL'),
                                          'email'  => (($club_app_users_state !== 'club_app_users_state_delete') ? $member->keycloakemail:'NULL'),
                                          'state'  => $iModState) );
                }
                else {
                    if( isset($return) ) {
                        $aStatistics['ERR-Process-Keycloak-Failed'][$Kunde["mioid"]][] = array('RETURN' => $return);
                    }
                    else{
                        $aStatistics['ERR-Process-Keycloak-Failed'][$Kunde["mioid"]][] = array('dummy' => 1);
                    }
                }
            }

            if( empty( $aStatistics['ERR-Process-Keycloak-Failed'][$Kunde["mioid"]] ) &&
               $club_app_users_state == 'club_app_users_state_enable' ) {
               $this->vvDeleteInfoByType($Kunde["mioid"], Information::OPHINEX_INFO_TYPE_APP_USERS_STATE);
            }
        }
    }

    /**
     * Immediate action: Process only Keycloak enabled state for a single club.
     * Does not connect to the Keycloak API, only processes the data in the database.
     *
     * @param int $iKundeId
     * @param string $strOperation
     * @return int|void
     * @throws KeycloakException
     */
    public function synchronizeNow(int $iKundeId, string $strOperation)
    {
        $aStatistics = $this->defineStatistics(config('constants.keycloakSync.MODE_DAILY'));
        $aStates = $this->defineStates(config('constants.keycloakSync.MODE_DAILY'));

        if (!Keycloak::setConfiguration(config('keycloak_api.keyCloakConfig'))['ok']) {
            return -1;
        }

        Keycloak::setTime((new DateTime('now'))->format('Y.m.d.H.i.s'));

        $Kunde = $this->vvClubData( $iKundeId );
        if( empty($Kunde) || !is_array($Kunde) || count($Kunde) == 0 ) {
            // No data
            return -1;
        }

        //Log::debug( "Op: $strOperation, Kunde in SyncNow ". print_r($Kunde, true) );

        // Only process clubs with app or app light, unless when enabling
        if( $strOperation != 'ENABLE' &&
            !$this->ProductVariantHas( $Kunde["PRODUCTVARIANT"], config('constants.VV.PRODUCT_APP') ) &&
            !$this->ProductVariantHas( $Kunde["PRODUCTVARIANT"], config('constants.VV.PRODUCT_APP_LIGHT') ) ) {
            return -1;
        }

        //Log::debug( "Kunde {$iKundeId} {$Kunde["NAME"]}" );

        $aStatistics['ERR-Process-Keycloak-Failed'][$iKundeId] = array();

        switch($strOperation) {
           case 'ENABLE':
               $bOnlyActive = true;  // Only active members are enabled
               $club_app_users_state = 'club_app_users_state_enable';
               break;
           case 'DISABLE':
               $bOnlyActive = false;  // Disable all members, regardless of Austrittsdatum
               $club_app_users_state = 'club_app_users_state_disable';
               break;
           case 'DELETE':
               $bOnlyActive = false;  // Delete all members, regardless of Austrittsdatum
               $club_app_users_state = 'club_app_users_state_delete';
               break;
            default:
                // Nothing to do
                return 0;
        }

        try {
            $oMembers = $this->vvGetMembersWithKCState( $iKundeId, $aStates[$club_app_users_state], $bOnlyActive);
            if (empty($oMembers)) {
                return 0;
            }
        }
        catch( \Exception $e ) {
            Log::error( "Error getting members for club {$iKundeId} {$Kunde["NAME"]}: " . $e->getMessage() );
            return -1;
        }

        foreach ($oMembers as $member) {
            if (!isset($member->keycloakuserid) || trim($member->keycloakuserid) === '') {
                continue;
            }

            switch($club_app_users_state) {
                case 'club_app_users_state_enable':
                    //THE STATE OF APP LICENSING IS SET TO ENABLED
                    $iModState = -1;
                    switch($member->keycloakstate) {
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED'):
                            $iModState = config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED');
                            break;
                    }
                    if( $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED') ||
                        $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED') ){
                        $return = Keycloak::enableUser(array('id' => $member->keycloakuserid));
                    }
                    break;

                case 'club_app_users_state_disable':
                    //THE STATE OF APP LICENSING IS SET TO DISABLED
                    $iModState = -1;
                    switch($member->keycloakstate) {
                        case config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED');
                            break;
                    }

                    if( $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED') ||
                        $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED') ) {
                        $return = Keycloak::disableUser(array('id' => $member->keycloakuserid));
                    }
                    break;

                case 'club_app_users_state_delete':
                    //THE STATE OF APP LICENSING IS SET TO DELETED
                    $iModState = -1;
                    switch($member->keycloakstate) {
                        case config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_EMAILVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_DISABLED_EMAILNOTVERIFIED');
                            break;
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED'):
                        case config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED'):
                            $iModState = config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DELETED_DISABLED_EMAILVERIFIED');
                            break;
                    }
                    if( $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILNOTVERIFIED')          ||
                        $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILVERIFIED')             ||
                        $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILNOTVERIFIED') ||
                        $iModState == config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILVERIFIED') ){
                        $return = Keycloak::deleteUser(array('id' => $member->keycloakuserid));
                    }
                    break;
            }
            if( ( $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILNOTVERIFIED')          &&
                  $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_EMAILVERIFIED')             &&
                  $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILNOTVERIFIED') &&
                  $iModState != config('constants.keycloakSync.TYP_APPEINLADSTATE_APPDELETED_DISABLED_EMAILVERIFIED') ) ||
                ( isset($return) && is_array($return) && isset($return['ok']) && $return['ok'] === config('constants.keycloakSync.KEYCLOAKAPI_SUCCESS') )
              ) {
                $this->vvSaveKCData($iKundeId, $member->adrid,
                               array( 'userid' => (($club_app_users_state !== 'club_app_users_state_delete') ? $member->keycloakuserid:'NULL'),
                                      'email'  => (($club_app_users_state !== 'club_app_users_state_delete') ? $member->keycloakemail:'NULL'),
                                      'state'  => $iModState) );
            }
        }

        //Log::debug( "Result Stats", $aStatistics );
    }

    /**
     * Synchronizes Keycloak user data with the local database on an hourly basis.
     * Basically updates the local database with the current state of the users in Keycloak
     * (so accesses Keycloak only in read mode).
     *
     * This function performs the following steps:
     * 1. Retrieves all clubs from the database.
     * 2. Iterates through each club and processes only those with app or app light.
     * 3. Retrieves members and updates their Keycloak status based on the defined states.
     * 4. Logs errors and statistics for the synchronization process.
     *
     * In order to avoid overloading the Keycloak server, the function waits for a few seconds after each
     * batch of requests.
     *
     * @param int $iDatabaseId The ID of the database to synchronize.
     * @return int|void Returns -1 on configuration failure, 1 if no data is found, or void on successful synchronization.
     * @throws KeycloakException If there is an error during the synchronization process.
     */
    public function synchronizeHourly(int $iDatabaseId)
    {
        $aStates = $this->defineStates( config('constants.keycloakSync.MODE_HOURLY') );

        if (!Keycloak::setConfiguration( config('keycloak_api.keyCloakConfig'))['ok']) {
            return -1;
        }

        Keycloak::setTime((new DateTime('now'))->format('Y.m.d.H.i.s'));

        $oAllClubs = $this->vvGetAllClubs();
        if( empty($oAllClubs) ) {
            // No data
            return 1;
        }

        foreach ($oAllClubs as $Kunde) {
            // Only process clubs with app or app light
            if( !$this->ProductVariantHas( $Kunde["productvariant"], config('constants.VV.PRODUCT_APP') ) &&
                !$this->ProductVariantHas( $Kunde["productvariant"], config('constants.VV.PRODUCT_APP_LIGHT') ) ) {
                continue;
            }

            try {
                $oMembers = $this->vvGetMembersWithKCState( $Kunde["mioid"], $aStates );
                if (empty($oMembers)) {
                    //Log::debug("No members found for club {$Kunde["MIOID"]} {$Kunde["NAME"]} {$Kunde["SHORTKEY"]}");
                    continue;
                }
            }
            catch( \Exception $e ) {
                Log::error( "Error getting members for club {$Kunde["mioid"]} {$Kunde["name"]}: " . $e->getMessage() );
                continue;
            }

            // Reset Statistics
            $aStatistics = $this->defineStatistics( config('constants.keycloakSync.MODE_HOURLY') );

            //Log::debug( "Kunde {$Kunde["mioid"]} {$Kunde["name"]}, Anz. Members ". count($oMembers) );

            $iCounter = 0;
            foreach ($oMembers as $oMember) {

                if( $iCounter % Keycloak::KEYCLOAKAPI_BATCH_SIZE == 0 ) {
                  // Wait for a few seconds before sending more requests to Keycloak
                  sleep(Keycloak::KEYCLOAKAPI_BATCH_PAUSE );
                }

                $response = Keycloak::verifyUser($iDatabaseId,
                    $Kunde["mioid"],
                    array(
                        'search' => 'userid',
                        'userid' => $oMember->keycloakuserid,
                        'email' => mb_strtolower($oMember->keycloakemail),
                        'adrid' => $oMember->adrid
                    ));
                $iCounter++;

                $aTemp = array(
                   'DATA' =>  array(
                                'keycloakemail' => $oMember->keycloakemail,
                                'adrid' => $oMember->adrid,
                                'keycloakuserid' => $oMember->keycloakuserid,
                                'keycloakstate' => $oMember->keycloakstate,
                                'deleted' => false
                              ),
                );

                if( isset($response['RETURN']['response']) ) {
                    $aTemp['RETURN'] = $response['RETURN']['response'];
                }
                else {
                    $aTemp['RETURN'] = $response['RETURN'];
                }

                if( $response['STATISTIC'] == 'ERR-Process-User-Not-Found-by-ID' ) {
                    $aTemp['DATA']['deleted'] = true;
                }

                $aStatistics['INF-Process-User-Status'][] = $aTemp;
            }

            foreach ($aStatistics['INF-Process-User-Status'] as $aData) {
                $enabled = ($aData['RETURN']['enabled'] ?? false);
                $emailVerified = ($aData['RETURN']['emailVerified'] ?? false);
                $deleted = $aData['DATA']['deleted'];
                $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED');
                if( $deleted ) {
                    // Delete records that have no equivalent anymore in Keycloak
                    if( $this->vvDeleteKCData($Kunde["mioid"], $aData['DATA']['adrid']) ) {
                        Log::debug( "Deleted record for club {$Kunde["mioid"]} {$Kunde["name"]} {$Kunde["shortkey"]} {$aData['DATA']['adrid']}");
                    }
                    else {
                        Log::debug( "Deleted failed for record for club {$Kunde["mioid"]} {$Kunde["name"]} {$Kunde["shortkey"]} {$aData['DATA']['adrid']}");
                    }
                    continue;
                }
                elseif( $enabled && !$emailVerified ) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED');
                }
                elseif( !$enabled && !$emailVerified ) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED');
                }
                elseif( !$enabled && $emailVerified ) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED');
                }
                elseif( $enabled && $emailVerified ) {
                    $status = config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED');
                }

                if ( $status != $aData["DATA"]['keycloakstate'] ) {
                    $this->vvSaveKCData( $Kunde["mioid"], $aData['DATA']['adrid'],
                                    array(  'userid' => $aData['RETURN']['id'],
                                            'email' => ($aData['DATA']['keycloakemail'] ?? $aData['RETURN']['username']),
                                            'user' => ($aData['RETURN']['username'] ?? ""),
                                            'state' => $status) );
                }
            }

            //Log::debug( "Result Stats", $aStatistics );
        }
    }

   /**
     * @param $mode
     * @return array
     * @throws KeycloakException
     */
    private function defineStatistics($mode)
    {
        $aStatistics = array();

        switch ($mode) {
            case config('constants.keycloakSync.MODE_HOURLY'):
                $aStatistics['INF-Process-User-Active']       = array();
                $aStatistics['INF-Process-User-Status']       = array();
                $aStatistics['INF-Process-User-Re-Invited']   = array();
                $aStatistics['INF-Process-User-Status']       = array();
                $aStatistics['ERR-Process-updateUser-Failed'] = array();
                break;
            case config('constants.keycloakSync.MODE_DAILY'):
                $aStatistics['INF-Process-User-Status']       = array();
                $aStatistics['INF-Process-User-Updated']      = array();
                $aStatistics['INF-Process-Keycloak-State']    = array();
                $aStatistics['ERR-Process-updateUser-Failed'] = array();
                $aStatistics['ERR-Process-Keycloak-Failed']   = array();
                break;
            default:
                throw new KeycloakException( "Unknown stats mode: $mode" );
        }

        return $aStatistics;
    }

    /**
     * @param $mode
     * @return array
     * @throws KeycloakException
     */
    private function defineStates($mode)
    {
        switch ($mode) {
            case config('constants.keycloakSync.MODE_HOURLY'):
                $aStates = array(
                    config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'),
                    config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'),
                    config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'),
                    config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED')
                );
                break;
            case config('constants.keycloakSync.MODE_DAILY'):
                $aStates = array(
                    'club_app_users_state_none' => array(
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED')
                    ),

                    'club_app_users_state_enable' => array(
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED')
                    ),

                    'club_app_users_state_disable' => array(
                        config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED')
                    ),

                    'club_app_users_state_delete' => array(
                        config('constants.keycloakSync.KEYCLOAKAPI_STATE_NOTINVITED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DISABLED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_DELETED_DISABLED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DISABLED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_EMAILVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILNOTVERIFIED'),
                        config('constants.keycloakSync.TYP_APPEINLADSTATE_APPSUSPENDED_DELETED_DISABLED_EMAILVERIFIED')
                    )
                );
                break;
            default:
                throw new KeycloakException( "Unknown state mode: $mode" );

        }

        return $aStates;
    }
}
