<?php
namespace App\Services;

use App\Exceptions\KeycloakException;
use App\Models\Databases;
use DateTime;
use Illuminate\Support\Facades\Log;
use App\Services\Keycloak;

/**
 * Class VVInterface
 * @package App\Services
 */
class VVInterface
{
    protected $mainApp;
    protected $aFileFirma;
    protected $strHtdocsPath;
    protected $aTexts;
    protected $iKundeId;
    protected $aConfigFlags;
    protected $conn;
    private $iDatabaseId;
    private $iClubId;

    /**
     * VVInterface constructor.
     * @param integer $database_id Database ID
     * @param integer|null $p_iKundeId Optional Customer ID
     */
    public function __construct( $database_id = null, $p_iKundeId = null )
    {
        global $mainApp, $aFileFirma, $aAdminDefaults, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;

        $mainApp = null;
        $conn = null;

        $this->iDatabaseId = $database_id;
        $p_aAltDBData = null;

        if( !is_null($database_id) ) {
            // Get database data from database_id
            $p_aAltDBData = Databases::getIfexists($database_id);
            if ($p_aAltDBData === FALSE) {
                abort(400, 'Invalid database id');
            }
        }

        $g_strMioHtdocsPath = env('VV_INCLUDE_PATH')."/htdocs";

        require_once( $g_strMioHtdocsPath . "/mio_db_types.php" );
        require_once( $g_strMioHtdocsPath . "/main_include.php" );
        require_once( $g_strMioHtdocsPath . "/mio_color_types.php" );

        require_once( $g_strMioHtdocsPath . "/classes/OphinexApplication.php" );

        // Load and apply database host mapping
        $dbmapJson = env("DB_HOST_MAP");
        if( !empty($dbmapJson) && file_exists(base_path($dbmapJson)) ) {
            $jsonFile = file_get_contents(base_path($dbmapJson));
            $json = json_decode($jsonFile, true);
            if( !empty($json) ) {
                $aDBMap = $json;
                if( array_key_exists($p_aAltDBData["dbhost"], $aDBMap) ) {
                    // Replace dbhost with mapped value
                    $p_aAltDBData["dbhost"] = $aDBMap[$p_aAltDBData["dbhost"]];
                }
            }
        }

        $mainApp = new \OphinexApplication(FALSE, FALSE, 0, $p_aAltDBData);  // Do general initialization

        $mainApp->loadLib( "/global_require.php" );
        $mainApp->loadLibs( array('/lib/tipmio_functions2.php',
                                  '/lib/adodb/adodb-time.inc.php',
                                  '/classes/unixtimestamp.php',
                                  '/classes/tabledata_fileinfo.php',
                                  '/classes/largefile.php',
                                  '/mio/lib/lib_largefiles.php',
                                  '/mio/lib/si_menuitem.lib.php',
                                  '/lib/si_banners.php',
                                  '/classes/lobcontainer.php',
                                  '/classes/dbconnection.php',
                                  '/classes/db/sqlparser.php',
                                  '/classes/db/stmtpreparer.php',
                                  '/classes/listrecords.php',
                                  '/classes/sessionhandler.php',
                                  '/classes/sqltools.php',
                                  '/mio/lib/si_suchelisteben.php',
                                  '/lib/patchwork/vendor/autoload.php',
                                  '/lib/si_database.php',
                                  '/lib/check_functions.php',
                                  '/lib/skinfunctions.php',
                                  '/lib/tipmio_functions.php',
                                  '/lib/mio_functions_sql.php',
                                  '/lib/image_functions.php',
                                  '/mio/lib/si_bankdatencheck.php',
                                  '/mio/lib/si_adressen.php',
                                  '/lib/si_information.php') );

        if( !empty($p_iKundeId) ) {
            $iKundeId = (int)$p_iKundeId;
            $this->iClubId = $iKundeId;
        }
        else {
            $this->iClubId = null;
        }

        \Patchwork\Utf8\Bootup::initAll(); // Enables the portablity layer and configures PHP for UTF-8
        \Patchwork\Utf8\Bootup::filterRequestUri(); // Redirects to an UTF-8 encoded URL if it's not already the case
        \Patchwork\Utf8\Bootup::filterRequestInputs(); // Normalizes HTTP inputs to UTF-8 NFC

        // For local testing, overwrite db data from env file
        if( !is_null($p_aAltDBData) && env("LOCAL_TEST_ORACLE_DBHOST", '') !== '' ) {
            $p_aAltDBData["dbhost"] = env("LOCAL_TEST_ORACLE_DBHOST" );
            $p_aAltDBData["dbport"] = env("LOCAL_TEST_ORACLE_DBPORT");
            $p_aAltDBData["dbname"] = env( "LOCAL_TEST_ORACLE_DBNAME" );
            $p_aAltDBData["dbuser"] = env( "LOCAL_TEST_ORACLE_DBUSER" );
            $p_aAltDBData["dbuserpassword"] = env( "LOCAL_TEST_ORACLE_DBPASS" );
            $p_aAltDBData["usefilesystem"] = env( "LOCAL_TEST_ORACLE_USEFS" );
            $p_aAltDBData["active"] = 1;
            $p_aAltDBData["filearchiv"] = "qacopy";
        }

        // Connect to database
        if( !empty($p_aAltDBData) ) {
            $mainApp->initDB(FALSE, $p_aAltDBData);    // Connect to database
        }

        $mainApp->initializeDesign();

        $conn = $mainApp->getDBConnId();  // Retrieve DB connection ID
        if( empty($conn) ) {
            Log::error( "Fail: No DB connection" );
        }

        // Check if club exists
        if( !empty($iKundeId) && !VereinExists( $conn, $iKundeId ) ) {
            abort( 400, "Invalid club id, $iKundeId" );
        }

        // Load language file of admin folder
        if( empty($aFileFirma) ) {
            $aFileFirma = array();
            $aFileFirma["LangFileName"] = "lang.de";
        }

        $aTexts = $mainApp->readLocalLanguageFile($g_strMioHtdocsPath . "/mio/admin/" . $aFileFirma["LangFileName"]);

        if( !empty($p_iKundeId) ) {
            $iKundeId = $p_iKundeId;
        }

        $this->saveGlobals();
    }

    /**
     * Destructor that clears the data.
     */
    public function __destruct()
    {
        $this->clearGlobals();
    }

    /**
     * Get the database ID
     * @return int
     */
    public function getDatabaseId()
    {
        return $this->iDatabaseId;
    }

    /**
     * Get the club ID
     * @return int|null
     */
    public function getClubId()
    {
        return $this->iClubId;
    }

    /**
     *
     */
    protected function saveGlobals()
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;

        $this->mainApp = $mainApp;
        $this->aFileFirma = $aFileFirma;
        $this->strHtdocsPath = $g_strMioHtdocsPath;
        $this->aTexts = $aTexts;
        $this->iKundeId = $iKundeId;
        $this->aConfigFlags = $g_aConfigFlags;
        $this->conn = $conn;
    }

    /**
     *
     */
    protected function loadGlobals()
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;

        $mainApp = $this->mainApp;
        $aFileFirma = $this->aFileFirma;
        $g_strMioHtdocsPath = $this->strHtdocsPath;
        $aTexts = $this->aTexts;
        $iKundeId = $this->iKundeId;
        $g_aConfigFlags = $this->aConfigFlags;
        $conn = $this->conn;
    }

    /**
     *
     */
    protected function clearGlobals()
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;

        unset( $mainApp );
        unset( $aFileFirma );
        unset( $g_strMioHtdocsPath );
        unset( $aTexts );
        unset( $iKundeId );
        unset( $g_aConfigFlags );
        unset( $conn );
        // Unset variables from OphinexApplication
        unset( $g_aLargeFile );
        unset( $OphId );
        unset( $aDBData, $iBenutzerId, $iKundeId, $iVisitorId );
        unset( $g_strVersionInfo, $g_MailProtocols, $g_aDFBnetPostfach, $g_strHelpUrl, $g_strLoginFiles );
        unset( $g_aSMTPData, $g_aIbanBicWS, $g_strProductType, $g_iBgProcDelInterval, $g_bEnableTracking );
        unset( $g_strEnvironment, $g_strAdvPdfSecret );
    }

    /**
     * @param $shortkey
     * @return false
     */
    public function getMIOIdByShortkey($shortkey)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();

        $aKundeData = GetKundeByStrKey( $conn, $shortkey );
        if( !empty($aKundeData) ) {
            return $aKundeData["MIOId"];
        }
        else {
            return FALSE;
        }
    }

   /**
     * @return array|array[]
     */
    public function vvGetCountryList( $p_bUnchanged = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib('/lib/si_laender.php');

        $aData = GetLaenderList( $this->conn, "de" );

        if( !$p_bUnchanged ) {
            $aData = array_map(function ($tag) {
                return array(
                    'id' => $tag['Id'],
                    'value' => $tag['Name']
                );
            }, $aData);

            return array_values($aData);
        }
        else {
            return array_map(function ($tag) {
                return array(
                    'id' => $tag['Id'],
                    'value' => $tag['Name'],
                    'code' => $tag['Code']
                );
            }, $aData);
        }
    }

    /**
     * @param $member_id
     * @return mixed
     */
    public function vvGetMitgliedsData($member_id)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/si_keycloak.php',
            '/mio/lib/api/vv_api_adressen.php',
            '/mio/lib/si_adressenzusatzdaten.php',
        ));

        $aData = apiGetMitgliedsData( $this->conn, $this->iKundeId, $member_id );
        return $aData;
    }

    /**
     * @param $member_id
     * @param $mime_type
     * @return mixed
     */
    public function vvGetMitgliedsImage($member_id, &$mime_type)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/api/vv_api_adressen.php',
        ));

        return apiGetMitgliedImage( $this->conn, $this->iKundeId, $member_id, $mime_type );
    }

    /**
     * @param $member_id
     * @return mixed
     */
    public function vvGetProfileData($member_id)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/api/vv_api_adressen.php',
            '/mio/lib/si_adressenzusatzdaten.php',
        ));

        $aData = apiGetProfileData( $this->conn, $this->iKundeId, $member_id );
        return $aData;
    }

    /**
     * @return mixed
     */
    public function vvGetLetterSalutations( $p_bUnchanged = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/functions_wertelisten.php');

        $aData = GetAllBriefanrede( $this->conn, $this->iKundeId );

        if( empty($aData) ) {
            return array();
        }

        $aData = array_map(function($tag) {
            return array(
                'id' => $tag['BriefanredeId'],
                'value' => $tag['BriefanredeName']
            );
        }, $aData);

        if( !$p_bUnchanged ) {
            return array_values($aData);
        }
        else {
            return $aData;
        }
    }

    /**
     * @return array
     */
    public function vvGetMaritalStatus( $p_bUnchanged = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/functions_wertelisten.php');

        $aData = GetAllFamilien( $this->conn, $this->iKundeId );

        if( empty($aData) ) {
            return array();
        }

        $aData = array_map(function($tag) {
            return array(
                'id' => $tag['FamilienstandId'],
                'value' => $tag['FamilienstandName']
            );
        }, $aData);

        if( !$p_bUnchanged ) {
            return array_values($aData);
        }
        else {
            return $aData;
        }
    }

    /**
     * @return array
     */
    public function vvGetSalutations( $p_bUnchanged = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/functions_wertelisten.php');

        $aData = GetAllAnrede( $this->conn, $this->iKundeId );

        $aData = array_map(function($tag) {
            return array(
                'id' => $tag['AnredeId'],
                'value' => $tag['AnredeName']
            );
        }, $aData);

        if( !$p_bUnchanged ) {
            return array_values($aData);
        }
        else {
            return $aData;
        }
    }

    /**
     * @return array
     */
    public function vvGetTitles( $p_bUnchanged = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/functions_wertelisten.php');

        $aData = GetAllTitel( $this->conn, $this->iKundeId );

        if( empty($aData) ) {
            return array();
        }

        $aData = array_map(function ($tag) {
            return array(
                'id' => $tag['TitelId'],
                'value' => $tag['TitelName']
            );
        }, $aData);

        if( !$p_bUnchanged ) {
            return array_values($aData);
        }
        else {
            return $aData;
        }
    }

    /**
     * @return array
     */
    public function vvCheckBankdata( $p_strIBAN, $p_bAllData = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib('/lib/check_functions.php');
        $mainApp->loadClass( '/classes/ccibanbic.class.php' );

        $aCheckedBankData = CheckBankdata( 'Max Mustermann',
                                           '',
                                           '',
                                           '',
                                           $p_strIBAN,
                                           '' );

        if( $aCheckedBankData["Status"] != BANKCHK_STAT_CORRECTED && $aCheckedBankData["Status"] != BANKCHK_STAT_OK ) {
            $aData = array( "status" => "fail",
                            "msg" => $aCheckedBankData["Error"] );
        }
        else {
            $aData = array( "status" => "success" );

            // Get corrected data
            $aData["data"]["name"] = $aCheckedBankData["Data"]["Institut"];
            $aData["data"]["iban"] = $aCheckedBankData["Data"]["IBAN"];
            $aData["data"]["bic"]  = $aCheckedBankData["Data"]["BIC"];

            if( $p_bAllData ) {
                $aData["Data"]["account"] = $aCheckedBankData["Data"]["Kontonummer"];
                $aData["Data"]["bankcode"] = $aCheckedBankData["Data"]["BLZ"];
                $aData["Data"]["countrycode"] = $aCheckedBankData["Data"]["CountryCode"];
            }
        }

        return $aData;
    }

    /**
     * @param $member_id
     * @return mixed
     */
    public function vvGetAbtBeitraege($member_id)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib( '/mio/lib/api/vv_api_adressen.php' );

        $aData = apiGetAbteilBeitragInfo( $this->conn, $this->iKundeId, $member_id );
        return $aData;
    }

    /**
     * @param $database_id
     * @return array
     */
    public function vvGetClubData($database_id)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLib( '/mio/lib/si_vereinsdaten.php' );

        $aData = GetKundeData( $this->conn, $this->iKundeId );
        $aVereinsData = GetVereinsdaten($this->conn, $this->iKundeId);

        if( empty($aData) || empty($aVereinsData) ) {
            return array();
        }

        $aFinal = array();
        $aFinal["name"] = $aData["Name"];
        $aFinal["street"] = $aData["Strasse"];
        $aFinal["street2"] = $aData["Strasse2"];
        $aFinal["postalcode"] = $aData["PLZ"];
        $aFinal["city"] = $aData["Ort"];
        $aFinal["state"] = $aData["Bundesland"];
        $aFinal["country"] = $aData["Land"];
        $aFinal["email"] = $aData["Mail"];
        $aFinal["phone"] = $aData["Telefon"];
        $aFinal["homepage"] = $aData["Homepage"];
        $aFinal["facebookLink"] = $aVereinsData["FacebookLink"];
        $aFinal["twitterLink"] = $aVereinsData["TwitterLink"];
        $aFinal["instagramLink"] = $aVereinsData["InstagramLink"];
        $aFinal["linkToVV"] = env("VV_URL", "") . DIRECTORY_SEPARATOR . "login" . DIRECTORY_SEPARATOR . $aData["Shortkey"];
        $aFinal["linkToFIBU"] = env("FIBU_URL", "") . $aData["Shortkey"];

        if( !empty($aData["LogoId"]) ) {
          $aFinal["logourl"] = env("VV_URL") . "/getlogo.php?id=" . \OphinexApplication::_simpleScramble($database_id . "-" . $this->iKundeId . "-" . $aData["LogoId"], TIPMIO_SCRAMBLE_CONST);
        }
        else {
          $aFinal["logourl"] = "";
        }

        return $aFinal;
    }

    /**
     * @param $member_id
     * @param $data
     * @return integer Error code: 0=OK, 1=Member not found, 2=Creates duplicate, 3=Error on save
     * @throws \Exception
     */
    public function vvUpdateMemberInfo($member_id, $data)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/api/vv_api_adressen.php',
            '/mio/lib/si_anderungshistorie.php',
            '/mio/lib/si_changerequestdata.php',
            '/mio/lib/si_adressenzusatzdaten.php',
            '/mio/lib/si_vereine.php',
        ));

        // Check if requested change would create a duplicate member
        if( !empty($data["firstName"]) && !empty($data["lastName"]) && !empty($data["street"]) && !empty($data["city"]) &&
            ExistAddress($this->conn, $this->iKundeId, $data["firstName"], $data["lastName"], $data["street"], $data["city"], $member_id, TYP_ADRESS_Mitarb) ) {
            return 2;
        }

        // Birthday sharing option does not go through change request
        if( isset($data["shareBirthday"]) ) {
            UpdateBenutzerClubShareBday( $this->conn, $this->iKundeId, $member_id, $data["shareBirthday"], TYP_ANDHISTSRC_App );
        }

        $bResult = apiRegisterChangeRequest($this->conn, $this->iKundeId, $member_id, $data, 1);
        return ($bResult == TRUE ? 0 : 3);
    }

    /**
     * @param $member_id
     * @param $data
     * @return integer Error code: 0=OK, 1=Member not found, 3=Error on save
     * @throws \Exception
     */
    public function vvUpdateBankInfo($member_id, $data)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/api/vv_api_adressen.php',
            '/mio/lib/si_anderungshistorie.php',
            '/mio/lib/si_adressenzusatzdaten.php',
            '/mio/lib/si_changerequestdata.php',
            '/mio/lib/si_vereine.php',
        ) );

        $bResult = apiRegisterChangeRequest($this->conn, $this->iKundeId, $member_id, $data, 2);
        return ($bResult == TRUE ? 0 : 3);
    }

    /**
     * @param integer $member_id
     * @param string $data base64 Image file data
     * @return boolean TRUE on success, FALSE on error
     * @throws \Exception
     */
    public function vvSaveMemberPhoto($member_id, $mime_type, $data)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/api/vv_api_adressen.php',
            '/mio/lib/si_vereine.php'
        ) );

        $bResult = apiPostMitgliedImage($this->conn, $this->iKundeId, $member_id, $mime_type, $data);
        return $bResult;
    }

    public function vvIsMemberInClub( $member_id )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/api/vv_api_adressen.php',
        ) );

        // Validate that AdrId exists and is part of this club
        return apiValidateAdrId( $this->conn, $member_id, $this->iKundeId );
    }

}
