<?php
namespace App\Services;

use App\Exceptions\KeycloakException;
use App\Http\Controllers\CRMApiController;
use App\Jobs\KeycloakSync;
use App\Models\Databases;
use App\Models\Delayed;
use App\Models\Login;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Log;

/**
 * Class VVInterfaceCRM
 * @package App\Services
 */
class VVInterfaceCRM extends VVInterface
{
    const VV_DEACT_MSG = "block_message.html";

    const ACTION_NAMES = [
        DelayedActions::ACTION_TYPE_DELETE => "deleteDate",
        DelayedActions::ACTION_TYPE_UPDOWNGRADE => "updowngradeDate",
        DelayedActions::ACTION_TYPE_CANCEL => "cancelDate",
        DelayedActions::ACTION_TYPE_CHANGE => "changeStateDate",
    ];

    /**
     * VVInterfaceCRM constructor.
     * @param $database_id
     * @param $p_iKundeId
     */
    public function __construct( $database_id = null, $p_iKundeId = null )
    {
        parent::__construct( $database_id, $p_iKundeId );
    }

    /**
     * Destructor that clears the data.
     */
    public function __destruct()
    {
        parent::__destruct();
    }

    /**
     * Disable or enable a club completely, in VV DB and CAdmin.
     *
     * @param integer $p_iDatabaseId Database Id
     * @param integer $p_iClubId Unique customer identifier
     * @param boolean $p_bEnable true to enable, false to disable
     * @return void
     */
    public function disableEnableClub( $p_iDatabaseId, $p_iClubId, $p_bEnable )
    {
        global $mainApp, $conn;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/api/crm_api_clubs.php');

        // Set Active flag for CAdmin Login table
        if( $p_bEnable ) {
            // Enable
            $shortkey = GetKundeShortkey($conn, $p_iClubId);
            if( !empty($shortkey) ) {
                Login::where('shortkey', $shortkey)
                    ->update([
                        'active' => 1,
                        'activemessage' => ""
                    ]);
            }
        }
        else {
            // Disable
            $shortkey = GetKundeShortkey($conn, $p_iClubId);
            if( !empty($shortkey) ) {
                Login::where('shortkey', $shortkey)
                    ->update([
                        'active' => 0,
                        'activemessage' => self::VV_DEACT_MSG
                    ]);
            }
        }

        // Enable/disable in MIOKUNDEN
        DisableEnableClub( $conn, $p_iClubId, $p_bEnable );

        // Enable or disable Keycloak users of this club, if any
        dispatch( new KeycloakSync($p_iDatabaseId, $p_iClubId, ($p_bEnable ? 'enable' : 'disable')) )->onQueue('sync');
    }

    /**
     * Process the action date (delete, updowngrade, cancel, change state) fields from the API.
     *
     * If the date is given and not empty, the date will be stored and the action triggered on that date (when
     * the given day begins, so the given date is the first day where the state change is active. If the date is
     * in the past or today or omitted, the action is done immediately (triggered by false return value). If the
     * date is given but empty and a previous action date is present, the action is reverted and the date removed,
     * thus effectively nullifying the action request. If the date is given and not empty and a previous action
     * date is present, a 403 error is generated.
     *
     * @param integer $p_iDatabaseId Database Id
     * @param integer $p_iClubId Unique customer identifier
     * @param string $p_strActionDate Given action date, if any (empty or format Y-m-d)
     * @param integer $p_iType Type of action (one of DelayedActions::ACTION_TYPE_*)
     * @param boolean $p_bActive If true, the club is to be activated, if false, the club is to be deactivated.
     *    This is only relevant for $p_iType=ACTION_TYPE_CHANGE.
     *    For $p_iType=ACTION_TYPE_UPDOWNGRADE, this is the new state (true=upgrade, false=downgrade).
     * @return boolean true if action date was processed/stored, false if action should happen immediately.
     * @throws \Exception
     */
    private function processActionDate($p_iDatabaseId, $p_iClubId, $p_strActionDate, $p_iType, $p_bActive = null)
    {
        // See if there already is an action date stored
        $delayedAction = (new Delayed())->getAction($p_iDatabaseId, $p_iClubId, $p_iType);

        if( !is_null($p_strActionDate) ) {
            $p_strActionDate = trim($p_strActionDate);
            if( empty($p_strActionDate) && !empty($delayedAction) ) {
                // Action was cancelled
                // Delete job
                Delayed::where('id', $delayedAction->id)->delete();
                return true;
            }
            elseif( !empty($p_strActionDate) && !empty($delayedAction) && !empty($delayedAction->action_date) ) {
                // Action date already set
                abort(403, self::ACTION_NAMES[$p_iType] . " already set");
            }
            else {
                $iActionDate = strtotime($p_strActionDate);
                if (empty($iActionDate)) {
                    // Date could not be parsed
                    abort(422, self::ACTION_NAMES[$p_iType] . " is invalid");
                }
                if ($iActionDate <= time()) {
                    // Date in the past, so execute action immediately
                    return false;
                }
                elseif ($iActionDate > time()) {
                    DelayedActions::create($p_iDatabaseId, $p_iClubId, $p_iType, $iActionDate, $p_bActive);
                    return true;
                }
            }
        }
        else {
            // No action date given, so execute action immediately
            return false;
        }

        return true;
    }

    /**
     * Obtain statistical data from the specified club.
     *
     * @param integer $club_id Unique customer identifier
     * @return array Array with data
     */
    public function vvGetClubStatistics( $club_id )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/mio/lib/si_vereinsdaten.php',
            '/mio/lib/api/crm_api_stats.php',
        ));

        $iKundeId = $club_id;

        $aData = array();
        $aData["totalMembers"] = GetCountActualMembers( $conn, $iKundeId );

        $aHomeStats = GetHomeUserStatistic( $conn, $iKundeId );
        $aData["activeMembers"] = $aHomeStats["iNumActivPartners"];
        $aData["formerMembers"] = $aHomeStats["iNumInactivPartners"];
        $aData["anonMembers"] = $aHomeStats["iNumAnons"];
        $aData["totalOthers"] = apiCRMGetTotalOthers( $conn, $iKundeId );
        $aData["totalRecords"] = $aData["totalOthers"] + $aData["totalMembers"];
        $aData["inactiveUsers"] = $aHomeStats["iNumInactivUsers"];
        $aData["normalUsers"] = ($aHomeStats["iNumUsers"] - $aHomeStats["iNumAdmins"]);  // NumUsers include the admins, so we must subtract those
        $aData["adminUsers"] = $aHomeStats["iNumAdmins"];
        $aData["lastLogin"] = apiCRMGetLastLogin( $conn, $iKundeId );
        $aData["clubType"] = apiCRMGetClubType( $conn, $iKundeId );

        return $aData;
    }

    /**
     * @return array
     */
    public function vvGetDesigns( $p_bUnchanged = FALSE )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLib('/lib/skinfunctions.php');

        $aData = GetAllSkins( $this->conn );

        $aData = array_map(function($tag) {
            return array(
                'id' => $tag['SkinId'],
                'value' => $tag['Name']
            );
        }, $aData);

        if( !$p_bUnchanged ) {
            return array_values($aData);
        }
        else {
            return $aData;
        }
    }

    /**
     * @param $input
     * @param $database_id
     * @param $strShortkey
     * @return mixed
     * @throws \Exception
     */
    public function vvCreateClub( $input, $database_id, $strShortkey )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/si_vereinsdaten.php');
        $mainApp->loadLib('/lib/si_information.php');
        $mainApp->loadLib('/lib/initialize.lib.php');

        $fibu = 0;
        $prodVar = 0;

        if( !empty($input["attributes"]) ) {
            $attributes = $input["attributes"];
            if( isset($attributes["fibuLight"]) &&
                ($attributes["fibuLight"] === true || strtolower($attributes["fibuLight"]) === "true") ) {
                $fibu = 1;
            }
            if( isset($attributes["fibuPremium"]) &&
                ($attributes["fibuPremium"] === true || strtolower($attributes["fibuPremium"]) === "true") ) {
                $prodVar += 100;
            }
            if( isset($attributes["small"]) &&
                ($attributes["small"] === true || strtolower($attributes["small"]) === "true") ) {
                $prodVar += 10;
            }
            // You can either have app or appLight, not both
            if( isset($attributes["app"]) &&
                ($attributes["app"] === true || strtolower($attributes["app"]) === "true") ) {
                $prodVar += 1;
            }
            elseif( isset($attributes["appLight"]) &&
                ($attributes["appLight"] === true || strtolower($attributes["appLight"]) === "true") ) {
                $prodVar += 2;
            }
        }

        // Prepare club creation
        $aData = array();
        $aData["Name"] = $input["clubName"];

        $aData["Land"] = (!empty($input["country"]) ? GetCountryName($conn, $input["country"], "de") : "");
        $aData["LandId"] = $input["country"] ?? "NULL";
        $aData["Ort"] = $input["city"] ?? "";
        $aData["Postfach"] = "";
        $aData["PLZ"] = $input["postalCode"] ?? "";
        $aData["Strasse"] = $input["street"] ?? "";
        $aData["Strasse2"] = "";
        $aData["Email"] = $input["email"];
        $aData["Fax"] = "";
        $aData["Handy"] = "";
        $aData["Homepage"] = "";
        $aData["Telefon"] = "";
        $aData["Bemerkung"] = "";
        $aData["KundenNr"] = $input["customerNumber"] ?? "";
        $aData["AnsprechPartner"] = (isset($input["contactFirstname"]) && isset($input["contactLastname"]) ? substr( $input["contactFirstname"] . " " . $input["contactLastname"], 0, 50) : "");
        $aData["FunktionPartner"] = "";
        $aData["ShortKey"] = $strShortkey;
        $aData["SkinId"] = $input["design"];
        $aData["LicensedUsers"] = $input["numberOfUsers"];
        $aData["Fibu"] = $fibu;
        $aData["ProductVariant"] = $prodVar;
        if( isset($input["agbCheck"]) && !empty($input["agbCheck"])) {
            $aData["AGBCheck"] = strtotime($input["agbCheck"]);
        }
        else {
            $aData["AGBCheck"] = '';
        }

        // Config flags
        $aData["ConfigFlags"] = 0;
        foreach ($g_aConfigFlags as $iFlagId => $strFlagName) {
          if (isset($$strFlagName)) {
            $aData["ConfigFlags"] = $aData["ConfigFlags"] | $iFlagId;
          }
        }

        // Create club
        $iKundeId = NewKunde( $conn, $aData );

        if (isset($iKundeId) && $iKundeId > 0) {
          $aVereinsData = array();
          $aVereinsData["UstPflicht"] = 0;
          // set "EingetragenerVerein" to 0 (No)
          $aVereinsData["EingetragenerVerein"] = 0;
          $aVereinsData["Geschaeftsjahr"] = 0;
          if( SaveVereinsdaten($conn, $iKundeId, $aVereinsData) == -1 ) {
            throw new \Exception( "Error saving Vereinsdaten" );
          }

          // Save trial period
          if( $input["testMode"] && !empty($input["testEndDate"]) ) {
            $aInfoData = array();
            $aInfoData["Type"] = OPHINEX_INFO_TYPE_LAUFZEIT_ENDE;
            $aInfoData["Label"] = "EndDate";
            $aInfoData["Content"] = date("d.m.Y", strtotime($input["testEndDate"]));
            $aInfoData["Sort"] = 0;
            if( InsertInformation($conn, $aInfoData, $iKundeId) == -1 ) {
              throw new \Exception( "Error saving Vereinsdaten" );
            }
          }

          SaveEinstell($conn, $iKundeId, TYP_ADMEINST_Language, TYP_LANGUAGE_GERMAN);
          SaveEinstell($conn, $iKundeId, TYP_ADMEINST_ShowAge, 0);

          // Initialize MenuItem table
          if (InitMenuItems($conn, $iKundeId) < 0) {
            throw new \Exception( "Failed to initialize menu items" );
          }

          // Initialization of several database tables
          if( !InitializeVereinData( $conn, $iKundeId, $g_strMioHtdocsPath, TYP_LANGUAGE_GERMAN ) ) {
            throw new \Exception( "Failed to initialize base data" );
          }
        }
        else {
          throw new \Exception( "Error creating Kunde" );
        }

        $aData = new \stdClass();
        $aData->databaseId = $database_id;
        $aData->clubId = $iKundeId;
        $aData->shortKey = $strShortkey;

        return $aData;
    }

   /**
     * @param $input
     * @param $iKundeId
     * @return array|string|string[]|null
     */
    public function vvCreateFirstUser( $input, $p_iKundeId )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/si_bbutler.php');
        $iKundeId = $p_iKundeId;

        $aReplace = array('/[^a-z0-9_\-]/i');   // Replace anything except alphanumeric characters and '_', '-'
        $strUsername = $input["clubName"];
        $strNewUsername = preg_replace($aReplace, '', $strUsername);
        if( strlen($strNewUsername) > 50 ) {
          $strNewUsername = substr($strNewUsername, 0, 50); // Max length: 50
        }

        // Make sure name and last name are always set for FIBU registration
        $strVorname = substr($input["contactFirstname"] ?? $input["clubName"], 0, 50);
        $strNachname = substr($input["contactLastname"] ?? "Testen", 0, 255);

        $strPasswd = env("SVEREIN_DFLT_ADMINPASS_PLAIN" );
        $strPassword = md5(PASS_PREFIX.$strPasswd);
        // insert Benutzer
        $aData  =  array();
        $aData["Admin"]       = 1 ;// admin user
        $aData["DemoUser"]    = 0;
        $aData["UserName"]    = $strNewUsername;
        $aData["Password"]    = $strPassword;
        $aData["Aktiv"]       = 1;
        $aData["Vorname"]     = $strVorname;
        $aData["Nachname"]    = $strNachname;
        $aData["Strasse" ]    = $input["strasse"] ?? "";
        $aData["Strasse2"]    = "";
        $aData["PLZ"]         = $input["postalCode"] ?? "";
        $aData["Ort"]         = $input["city"] ?? "";
        $aData["Land" ]       = (isset($input["country"]) ? GetCountryName($conn, $input["country"], "de") : "");
        $aData["Telefon"]     = "";
        $aData["Mobil"]       = "";
        $aData["Email"]       = $input["email"];

        $strDBError = "";
        $iNewBenutzerId = NewBenutzer( $conn, $iKundeId, $aData, "NULL", $strDBError );
        if( $iNewBenutzerId <= 0 ) {
          throw new \Exception( "Error creating user" );
        }
        else {
          // Set initial permissions for new user to All
          $alistx = GetAllMenuItems( $conn, $iKundeId );
          foreach( $alistx as $iDmyKeyx => $aMenuItemsx ) {
            $imenuId = $aMenuItemsx["ItemId"];
            $strErrorRet = "";
            $res = InsertBerchtigungenItem( $conn, $iNewBenutzerId, $imenuId, TYP_BERECHT_RdWrtDel, $iKundeId, $strErrorRet);
            if( $res <= 0 ) {
              throw new \Exception( "Error inserting permissions" );
            }
          }

          // Set FIBU user role
          if( !BBSaveUserRole($conn, $iKundeId, $iNewBenutzerId, TYP_BBFIBU_Master) ) {
            throw new \Exception( "Error inserting FIBU role" );
          }
        }

        return $strNewUsername;
    }

    /**
     * Implement delete of club.
     *
     * @param integer $p_iClubId Unique customer identifier.
     * @param integer $p_iDatabaseId Database Id (only used if $p_strDeleteDate not empty)
     * @param string $p_strDeleteDate Date of desired deletion (empty or format Y-m-d)
     * @return boolean TRUE on success, FALSE on failure.
     */
    public function vvDeleteClub( $p_iClubId, $p_iDatabaseId, $p_strDeleteDate = null )
    {
        $this->loadGlobals();

        // Process delete date
        if( !$this->processActionDate( $p_iDatabaseId, $p_iClubId, $p_strDeleteDate,
                                DelayedActions::ACTION_TYPE_DELETE ) ) {
            // Immediate delete
            return $this->deleteClub($p_iDatabaseId, $p_iClubId);
        }

        return true;
    }

    /**
     * @param $input
     * @param integer $database_id
     * @param integer $club_id
     * @return mixed
     * @throws \Exception
     */
    public function vvUpdateClub( $input, $database_id, $club_id )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLib('/lib/si_information.php');
        $mainApp->loadLib('/mio/lib/api/crm_api_stats.php');
        $mainApp->loadLib('/mio/lib/api/crm_api_clubs.php');

        $iKundeId = $club_id;
        $curProdVar = GetProduktVariante($conn, $iKundeId);

        // Prepare club creation
        $aData = $this->initClubData($conn, $input);

        // Process cancel date
        if( isset($input["cancelDate"]) &&
            !$this->processActionDate($database_id, $club_id, $input["cancelDate"],
                DelayedActions::ACTION_TYPE_CANCEL)) {
            // Immediate cancel
            $this->disableEnableClub($database_id, $club_id, false);
        }

        // Process changeState date
        if( isset($input["changeStateDate"]) &&
            $this->processActionDate($database_id, $club_id, $input["changeStateDate"],
                DelayedActions::ACTION_TYPE_CHANGE, $input["active"])) {
            // Process was successful, so state will be set later
            // Unset received state for now
            unset($input["active"]);
        }

        // Process updowngrade date
        if( isset($input["attributes"]["small"]) && isset($input["updowngradeDate"]) ) {
            $this->processUpdateDowngradeDate($input, $curProdVar, $database_id, $club_id);
        }

        // Process test end date FIBU
        if (isset($input["testEndDateFIBU"])) {
           $this->processTestEndDateFIBU($input["testEndDateFIBU"], $club_id);
        }
        // Process FIBU order
        // This is really a work-around so that testEndDateFIBU doesn't have to be implemented
        // in the CRM (see SV-249513).
        elseif (isset($input["attributes"]["fibuPremium"])) {
            $this->processFIBUOrder($input["attributes"]["fibuPremium"], $curProdVar, $club_id);
        }

        // Figure out the new setting for ProductVariant, based on the input attrivutes
        $prodVar = $curProdVar;
        if( !empty($input["attributes"]) ) {
          $fibu = null;
          $prodVar = DefineNewProduktVariante($input, $curProdVar, $fibu);
        }

        // If product variant has changed, process that
        if( $prodVar != $curProdVar ) {
            $aData["ProductVariant"] = $prodVar;
            $this->processAppEnableDisable($curProdVar, $prodVar, $database_id, $club_id);
        }

        if( isset($input["active"]) ) {
            $aData["Aktiv"] = ($input["active"] ? 1 : 0);
        }

        if( !empty($aData) && UpdateKunde( $conn, $iKundeId, $aData ) === FALSE ) {
            throw new \Exception("Error saving Kundendaten");
        }

        // Enable / disable club if necessary
        if( isset($aData["Aktiv"]) ) {
            $this->disableEnableClub( $database_id, $iKundeId, ($aData["Aktiv"] == 1) );
        }

        // Save trial period
        SaveTrialPeriod( $conn, $iKundeId, $input );

        return TRUE;
    }

    /**
     * @param $conn
     * @param $input
     * @return array
     */
    private function initClubData($conn, $input)
    {
        $aData = array();
        if( isset($input["clubName"]) ) {
            $aData["Name"] = $input["clubName"];
        }
        if( isset($input["country"]) ) {
            $aData["Land"] = GetCountryName($conn, $input["country"], "de");
            $aData["LandId"] = $input["country"];
        }
        if( isset($input["city"]) ) {
            $aData["Ort"] = $input["city"];
        }
        if( isset($input["postalCode"]) ) {
            $aData["PLZ"] = $input["postalCode"];
        }
        if( isset($input["street"]) ) {
            $aData["Strasse"] = $input["street"];
        }
        if( isset($input["email"]) ) {
            $aData["Email"] = $input["email"];
        }
        if( isset($input["customerNumber"]) ) {
            $aData["KundenNr"] = $input["customerNumber"];
        }
        if( isset($input["contactFirstname"]) || isset($input["contactLastname"]) ) {
            $aData["AnsprechPartner"] = substr( $input["contactFirstname"] . " " . $input["contactLastname"], 0, 50);
        }
        if( isset($input["design"]) ) {
            $aData["SkinId"] = $input["design"];
        }
        if( isset($input["numberOfUsers"]) ) {
            $aData["LicensedUsers"] = $input["numberOfUsers"];
        }
        if( isset($fibu) ) {
            $aData["Fibu"] = $fibu;
        }
        if( isset($input["agbCheck"]) ) {
            if( !empty($input["agbCheck"]) ) {
                $aData["AGBCheck"] = strtotime($input["agbCheck"]);
            }
            else {
                $aData["AGBCheck"] = '';
            }
        }

        return $aData;
    }

    /**
     * @param $curProdVar
     * @param $newProdVar
     * @param $database_id
     * @param $club_id
     * @return void
     */
    private function processAppEnableDisable($curProdVar, $newProdVar, $database_id, $club_id)
    {
        $this->loadGlobals();

        // Enabling App or AppLight
        if ((ProductVariantHas($newProdVar, VV_PRODUCT_APP) ||
                ProductVariantHas($newProdVar, VV_PRODUCT_APP_LIGHT)) &&
            (!ProductVariantHas($curProdVar, VV_PRODUCT_APP) &&
                !ProductVariantHas($curProdVar, VV_PRODUCT_APP_LIGHT))
        ) {
            $iEnable = 1;
        }
        elseif ((ProductVariantHas($curProdVar, VV_PRODUCT_APP) ||
                ProductVariantHas($curProdVar, VV_PRODUCT_APP_LIGHT)) &&
            (!ProductVariantHas($newProdVar, VV_PRODUCT_APP) &&
                !ProductVariantHas($newProdVar, VV_PRODUCT_APP_LIGHT))
        ) {
            // Disabling App or AppLight
            $iEnable = 0;
        }
        else {
            // Change doesn't affect App, don't change Keycloak users
            $iEnable = -1;
        }

        if ($iEnable != -1) {
            // Enable or disable Keycloak users of this club, if any
            dispatch( new KeycloakSync($database_id, $club_id, ($iEnable ? 'enable' : 'disable')) )->onQueue('sync');
        }
    }

    /**
     * @param $fibuPremium
     * @param $curProdVar
     * @param $club_id
     * @return void
     * @throws \Exception
     */
    private function processFIBUOrder($fibuPremium, $curProdVar, $club_id)
    {
        global $conn;

        $aCurrentInfo = GetInformationByType($conn, OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE, $club_id);
        if (!empty($aCurrentInfo) &&    // Current test period for FIBU exists
            ($fibuPremium === true || strtolower($fibuPremium) === "true") &&   // FIBU is being enabled
            ProductVariantHas($curProdVar, VV_PRODUCT_FIBU_PREMIUM) ) {   // even though it already is enabled
            // Club has test FIBU and already has FIBU but is being enabled again
            // This means the FIBU was ordered.
            // So we need to remove the test period
            if (!DeleteInformationByType($conn, $club_id, OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE)) {
                throw new \Exception("Error deleting testEndDateFIBU");
            }
        }
    }

    /**
     * @param $testEndDateFIBU
     * @param $club_id
     * @return void
     * @throws \Exception
     */
    private function processTestEndDateFIBU($testEndDateFIBU, $club_id)
    {
        global $conn;

        if (empty($testEndDateFIBU)) {
            // Remove test end date
            if (!DeleteInformationByType($conn, $club_id, OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE)) {
                throw new \Exception("Error deleting testEndDateFIBU");
            }
        }
        else {
            // Get existing value, if any
            $aCurrentInfo = GetInformationByType($conn, OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE, $club_id);
            $strNewContent = date("d.m.Y", strtotime($testEndDateFIBU));
            // Update trial period?
            if (!empty($aCurrentInfo)) {
                $aCurrentInfo = $aCurrentInfo[0]; // No more than one record per club possible
                Log::debug("Compare {$aCurrentInfo["Content"]} != $strNewContent ");
                if ($aCurrentInfo["Content"] != $strNewContent) {
                    $aCurrentInfo["Id"] = $aCurrentInfo["InfoId"];
                    $aCurrentInfo["Content"] = $strNewContent;
                    $aCurrentInfo["Type"] = OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE;
                    if (!UpdateInformation($conn, $aCurrentInfo, $club_id)) {
                        throw new \Exception("Error updating testEndDateFIBU");
                    }
                }
            }
            else {
                // Insert new trial period
                $aInfoData = array();
                $aInfoData["Type"] = OPHINEX_INFO_TYPE_FIBU_LAUFZEIT_ENDE;
                $aInfoData["Label"] = "FIBU Runtime";
                $aInfoData["Content"] = $strNewContent;
                $aInfoData["Sort"] = 0;
                if (InsertInformation($conn, $aInfoData, $club_id) == -1) {
                    throw new \Exception("Error saving testEndDateFIBU");
                }
            }
        }
    }

    /**
     * @param $input
     * @param $curProdVar
     * @param $database_id
     * @param $club_id
     * @return void
     * @throws \Exception
     */
    private function processUpdateDowngradeDate(&$input, $curProdVar, $database_id, $club_id)
    {
       $small = $input["attributes"]["small"];
       if( ((($small == true || strtolower($small) === "true")
                 && !ProductVariantHas($curProdVar, VV_PRODUCT_SMALL)) ||
             (($small == false || strtolower($small) === "false")
                 && ProductVariantHas($curProdVar, VV_PRODUCT_SMALL)))
       ) {
            if( $small == true || strtolower($small) === "true" ) {
                // Add SMALL
                $bUpgrade = false;
            }
            else {
                // Remove SMALL
                $bUpgrade = true;
            }

            if (!$this->processActionDate($database_id, $club_id, $input["updowngradeDate"],
                DelayedActions::ACTION_TYPE_UPDOWNGRADE, $bUpgrade)) {
                // Immediate updowngrade
                if( !$this->updowngradeClub($club_id, $bUpgrade) ) {
                    abort(500, "Up/Downgrade failed");
                }
            }
            else {
                // Process was successful, so new state will be set later
                unset( $input["attributes"]["small"] );
            }
        }

    }

    /**
     * @param $club_id
     * @return void
     */
    public function vvGetLoginInfo( $club_id )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();

        $iKundeId = $club_id;

        $aKundeData = GetKundeData( $conn, $iKundeId );
        if( empty($aKundeData) ) {
              abort(404, "Specified club not found");
        }
        $aData["link"] = env("VV_URL", "") . DIRECTORY_SEPARATOR . "login" . DIRECTORY_SEPARATOR . $aKundeData["Shortkey"];
        $aData["active"] = ($aKundeData["Aktiv"] ? TRUE : FALSE);
        $aData["shortkey"] = $aKundeData["Shortkey"];
        $aUsers = GetUserNames( $conn, $iKundeId, TRUE, TRUE );
        foreach( $aUsers as $aUser ) {
            $aClubUsers[] = array(
                "lastname" => $aUser["Nachname"],
                "firstname" => $aUser["Vorname"],
                "username" => $aUser["UserName"],
                "type" => ($aUser["SupportUser"] ? "support" : ($aUser["Admin"] ? "admin" : "user")),
                "email" => $aUser["Email"],
                "active" => ($aUser["Aktiv"] ? TRUE : FALSE)
            );
        }
        $aData["users"] = $aClubUsers;

        return $aData;
    }

    /**
     * @param $club_id
     * @param $subject
     * @return mixed
     */
    public function vvActivateSupportUser( $club_id, $subject )
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId, $g_aConfigFlags;
        $this->loadGlobals();
        $mainApp->loadLibs( array("/classes/phpMailerExtended.php") );

        $iKundeId = $club_id;
        $aTexts = $mainApp->readLocalLanguageFile( $g_strMioHtdocsPath . "/mio/admin/" . $aFileFirma["LangFileName"] );

        if (!empty($subject)) {
            $strBetreff = $subject;
        }
        else {
            $strBetreff = sprintf( $aTexts["DefaultSupportBetreff"], "CRM" );
        }

        $strMsgError = "";
        $aSupportUserData = array();
        if( !HandleSupportUser($conn, $iKundeId, $strBetreff, TRUE, $aSupportUserData, $strMsgError) ) {
            throw new \Exception("Error creating or activating support user: $strMsgError" );
        }

        $aData["user"] = "support";
        $aData["password"] = $aSupportUserData["Password"];

        return $aData;
    }

    /**
     * Checks if the specified club exists.
     *
     * @param integer $p_iClubId Unique customer/club identifier
     * @return mixed true if club exists, false otherwise
     */
    public function vvClubExists( $p_iClubId )
    {
        global $conn;
        $this->loadGlobals();

        return VereinExists( $conn, $p_iClubId );
    }

    /**
     * Delete the given club directly.
     *
     * @param integer $p_iDatabaseId ID of database.
     * @param integer $p_iClubId Unique customer identifier.
     * @return mixed
     */
    public function deleteClub( $p_iDatabaseId, $p_iClubId )
    {
        global $conn;
        $this->loadGlobals();
        if( !DeleteKunde( $conn, $p_iClubId ) ) {
            return false;
        }
        else {
            // Delete Keycloak users of this club, if any
            dispatch( new KeycloakSync($p_iDatabaseId, $p_iClubId, 'delete') )->onQueue('sync');
            return true;
        }
    }

    /**
     * Upgrade the given club to large or downgrades it to small.
     *
     * @param int $p_iClubId Unique customer identifier.
     * @param bool $p_bUpgrade true to upgrade, false to downgrade
     * @return bool true on success, false otherwise
     * @throws \Exception if the up/downgrade fails.
     */
    public function updowngradeClub($p_iClubId, $p_bUpgrade)
    {
        global $conn, $mainApp;
        $this->loadGlobals();
        $mainApp->loadLib('/mio/lib/api/crm_api_clubs.php');

        $bReturn = true;
        try {
            UpDowngradeClub($conn, $p_iClubId, $p_bUpgrade);
        }
        catch (\Exception $e) {
            Log::error( "Up/DowngradeClub failed: " . $e->getMessage() );
            $bReturn = false;
        }

        return $bReturn;
    }

    /**
     * @param $database_id
     * @return array
     */
    public function vvGetClubInfo($club_id)
    {
        global $mainApp, $aFileFirma, $g_strMioHtdocsPath, $aTexts, $conn, $iKundeId;
        $this->loadGlobals();
        $mainApp->loadLibs( array(
            '/lib/si_information.php',
            '/mio/lib/api/crm_api_clubs.php'
        ));

        $aData = GetCRMClubInfo( $this->conn, $club_id );

        if( empty($aData) ) {
            return array();
        }

        return $aData;
    }
}
