<?php

/** @var \Laravel\Lumen\Routing\Router $router */

/*
|--------------------------------------------------------------------------
| Application Routes
|--------------------------------------------------------------------------
|
| Here is where you can register all of the routes for an application.
| It is a breeze. Simply tell Lumen the URIs it should respond to
| and give it the Closure to call when that URI is requested.
|
*/

/**
 * User-related calls, require DB-Id, Club-Id and Member-Id
 */
$router->group(['middleware' => ['auth', 'validate_user'], 'prefix' => 'api'], function () use ($router) {
  /**
   * get getMemberInfo
   * Summary: Delivers stored address info on passed member
   * Notes: By passing in the database, club and member ids, you can obtain details about the member as currently stored in the verein.cloud web app.
   */
  $router->get('member-info', ['uses' => 'UsersApi@getMemberInfo']);

  /**
   * get getMemberPhoto
   * Summary: Delivers picture of the passed member, if available
   * Notes: By passing in the database, club and member ids, you can obtain a member photo if one is available.
   */
  $router->get('member-photo', ['uses' => 'UsersApi@getMemberPhoto']);

  /**
   * get getDeptsContribsInfo
   * Summary: Delivers information about the departments and contributions the member participates in
   * Notes: Obtain information about the departments that the member participates in and the contributions he/she is registered for.
   */
  $router->get('depts-and-contribs', ['uses' => 'UsersApi@getDeptsContribsInfo']);

  /**
   * post updateMember
   * Summary: Modifies member info
   * Notes: Modifies the data of the specified member. Only specify the fields that you want to update. All data fields are optional.
   */
  $router->post('member-info', ['uses' => 'UsersApi@updateMember']);

  /**
   * post uploadMemberPhoto
   * Summary: Allows upload of a member&#39;s photo.
   * Notes: Allow the binary file of a member&#39;s photo to the service. This will be stored in the member data.
   */
  $router->post('member-photo', 'UsersApi@uploadMemberPhoto');

  /**
   * post updateBankData
   * Summary: Modifies a member&#39;s bank data
   * Notes: Send a new IBAN for a member to the server so that a new request for bank data modification is created there.
   */
  $router->post('bank-info', ['uses' => 'UsersApi@updateBankData']);
});

/**
 * Club-based calls, require DB_id and Club_id only
 */
$router->group(['middleware' => ['auth','validate_club'], 'prefix' => 'api'], function () use ($router) {
  /**
   * get getProfileInfo
   * Summary: Delivers profile info about the passed member.
   * Notes:
   */
  $router->get('profile-info', ['uses' => 'UsersApi@getProfileInfo']);

  /**
   * get getProfilePhoto
   * Summary: Delivers picture of the passed member, if available
   * Notes: By passing in the database, club and member ids, you can obtain a member photo if one is available,
   * or a default image if none is available.
   */
  $router->get('profile-photo', ['uses' => 'UsersApi@getMemberPhoto']);

  /**
   * get getProfilePhotos
   * Summary: Delivers the pictures of the specified members in a batch.
   * Notes: By passing in the database, club and member ids, you can obtain the corresponding member photos if they are
   * available. For unavailable photos, a default image is returned.
   */
  $router->get('profile-photos', ['uses' => 'UsersApi@getMemberPhotos']);

  /**
   * get getListOfLetterSalutations
   * Summary: Return a list of all available letter salutations for the specified club.
   * Notes:
   */
  $router->get('list-letter-salutations', ['uses' => 'ClubsApi@getListOfLetterSalutations']);

  /**
   * get getListOfMaritalStatus
   * Summary: Return a list of all available marital status values for the specified club.
   * Notes:
   */
  $router->get('list-marital-status', ['uses' => 'ClubsApi@getListOfMaritalStatus']);

  /**
   * get getListOfSalutations
   * Summary: Return a list of all available salutations for the specified club.
   * Notes:
   */
  $router->get('list-salutations', ['uses' => 'ClubsApi@getListOfSalutations']);

  /**
   * get getListOfTitles
   * Summary: Return a list of all available titles for the specified club.
   * Notes:
   */
  $router->get('list-titles', ['uses' => 'ClubsApi@getListOfTitles']);

  /**
   * get club-info
   * Summary: Delivers information about the specified club
   * Notes:
   */
  $router->get('club-info', ['uses' => 'ClubsApi@getClubInfo']);
});

/**
 * DB calls, require only DB-IDs
 */
$router->group(['middleware' => ['auth', 'validate_db'], 'prefix' => 'api'], function () use ($router) {
  /**
   * get getListOfCountries
   * Summary: Return a list of all available countries in the VV system.
   * Notes: Only requires databaseId, not clubId.
   */
  $router->get('list-countries', ['uses' => 'DBApi@getListOfCountries']);
});

/**
 * Base calls, require no IDs
 */
$router->group(['middleware' => ['auth'], 'prefix' => 'api'], function () use ($router) {
  /**
   * get validateIBAN
   * Summary: Validate the passed IBAN and returns bank details if available
   * Notes:
   */
  $router->get('validate-iban', ['uses' => 'BaseApi@validateIBAN']);
});

/**
 * DB calls in CRM API, require only DB-IDs
 */
$router->group(['middleware' => ['auth', 'validate_crm_db'], 'prefix' => 'crm'], function () use ($router) {
    /**
     * get getLoginInfo
     * Summary: Delivers login info for the specified club.
     * Notes: By passing in the database and club id (or shortkey), you can obtain the login information for that club (available Administrator usernames and emails).
     */
    $router->get('login-info', 'CRMApiController@getLoginInfo');

    /**
     * get supportUser
     * Summary: Create, activate or return support user information for the given club.
     * Notes: Checks if the support user is already active for the given club.  If so, the password is reset and returned.
     *   If it exists but is inactive, it is activated, the password reset and the info returned.
     *   If it does not exist, it is created and activated and the info returned.
     */
    $router->get('support-user', 'CRMApiController@supportUser');

    /**
     * get club-info
     * Summary: Delivers information about the specified club
     * Notes:
     */
    $router->get('club-info', ['uses' => 'CRMApiController@getClubInfo']);

    /**
     * post createClub
     * Summary: Creates a new club in the specified database.
     * Notes: Create a new club in the specified database with the data from the parameters.
     */
    $router->post('create-club', 'CRMApiController@createClub');

    /**
     * get getListOfDesigns
     * Summary: Return a list of all available designs for the specified database.
     * Notes:
     */
    $router->get('list-designs', 'CRMApiController@getListOfDesigns');
});

/**
 * CRM Club-based calls, require DB_id and Club_id only
 */
$router->group(['middleware' => ['auth','validate_crm_club'], 'prefix' => 'crm'], function () use ($router) {
    /**
     * get getClubStats
     * Summary: Delivers statistics for the specified club.
     * Notes: By passing in the database and club ids, you can obtain statistics about the specified club. The info delivered is: - Number of records (relevant for SMALL) - Number of active members (\&quot;nicht ausgetreten\&quot;) - Number of former mambers (\&quot;ausgetreten\&quot;) - Number of anonymized members - Number of users - Number of admin users - Date and time of last login
     */
    $router->get('club-statistics', 'CRMApiController@getClubStats');

    /**
     * post deleteClub
     * Summary: Deletes the specified club from the specified database.
     * Notes: Delete the specified club from the database. This completely removes the club and all of its data in an irreversible process.
     */
    $router->post('delete-club', 'CRMApiController@deleteClub');

    /**
     * post updateClub
     * Summary: Updates data of an existing club in the specified database.
     * Notes: Update the specified data only for the fiven club in the given database.
     */
    $router->post('update-club', 'CRMApiController@updateClub');
});

/**
 * CRM Base calls, require no IDs
 */
$router->group(['middleware' => ['auth'], 'prefix' => 'crm'], function () use ($router) {
    /**
     * get getGeneralStats
     * Summary: Delivers statistics on all clubs in the system.
     * Notes: By calling this endpoint, you can obtain statistics about all clubs in all databases. The info delivered is:  Per database: - Total number of clubs - Number of active clubs - Number of inactive clubs - Number of clubs in test/trial mode - Type of database (MySQL/Oracle) - In case of Oracle DB: Current storage space  Totals (all databases): - Total number of clubs - Number of active clubs - Number of inactive clubs - Number of clubs in test/trial mode
     */
    $router->get('general-statistics', 'CRMApiController@getGeneralStats');

    /**
     * get getListOfDatabases
     * Summary: Return a list of all available databases.
     * Notes:
     */
    $router->get('list-databases', 'CRMApiController@getListOfDatabases');

    /**
     * get shortkeyAvailable
     * Summary: Checks if the given shortkey is available.
     * Notes: Checks the CAdmin database for availability of the given shortkey.
     */
    $router->get('shortkey-available', 'CRMApiController@shortkeyAvailable');

});

/**
 * BB calls, require no IDs
 */
$router->group(['middleware' => ['auth'], 'prefix' => 'bb'], function () use ($router) {
    /**
     * get validateUser
     * Summary: Validates a user through the Benutzer table of S-Verein.
     * Notes: Validate a user by checking if username and password are correct and match an existing user in the Benutzer table of S-Verein.
     */
    $router->get('validate-user', 'BBApi@validateUser');

    /**
     * get usersGet
     * Summary:
     * Notes:
     */
    $router->get('users', 'BBApi@usersGet');

    /**
     * get usersCountGet
     * Summary:
     * Notes:
     */
    $router->get('users/count', 'BBApi@usersCountGet');

    /**
     * get usersIdGet
     * Summary:
     * Notes:
     */
    $router->get('users/{id}', 'BBApi@usersIdGet');

   /**
     * post deliverClientCreds
     * Summary: Receives client credentials for the specified club, so that communication with the BB API is possible.
     * Notes: Called by BB after a successful login with the credentials needed for talking to the BuchhaltungsButler API.
     */
    $router->post('deliver-client-creds', 'BBApi@deliverClientCreds');
});

$controller = 'UsersApi@handleUnknown';
$router->get('/{route:.*}/', $controller);
$router->post('/{route:.*}/', $controller);
$router->put('/{route:.*}/', $controller);
$router->patch('/{route:.*}/', $controller);
$router->delete('/{route:.*}/', $controller);
